Package: chiavdf-hw
Version: {{ INSTALLER_VERSION }}
Architecture: {{ PLATFORM }}
Maintainer: Chia Network Inc <hello@chia.net>
Description: Hardware VDF Client
Depends: libgmp-dev, libboost-system-dev
